@@ -88,9 +88,7 @@ struct GdmSlavePrivate
         char            *parent_display_name;
         char            *parent_display_x11_authority_file;
 
-        /* user selected */
-        char            *selected_session;
-        char            *selected_language;
+        GArray          *display_x11_cookie;
 
         DBusGProxy      *display_proxy;
         DBusGConnection *connection;
@@ -191,18 +189,26 @@ get_script_environment (GdmSlave   *slave,
         g_hash_table_insert (hash, g_strdup ("PWD"), g_strdup ("/"));
         g_hash_table_insert (hash, g_strdup ("SHELL"), g_strdup ("/bin/sh"));
 
-        g_hash_table_insert (hash, g_strdup ("LOGNAME"), g_strdup (username));
-        g_hash_table_insert (hash, g_strdup ("USER"), g_strdup (username));
-        g_hash_table_insert (hash, g_strdup ("USERNAME"), g_strdup (username));
+        if (username != NULL) {
+                g_hash_table_insert (hash, g_strdup ("LOGNAME"),
+                                     g_strdup (username));
+                g_hash_table_insert (hash, g_strdup ("USER"),
+                                     g_strdup (username));
+                g_hash_table_insert (hash, g_strdup ("USERNAME"),
+                                     g_strdup (username));
 
-        pwent = getpwnam (username);
-        if (pwent != NULL) {
-                if (pwent->pw_dir != NULL && pwent->pw_dir[0] != '\0') {
-                        g_hash_table_insert (hash, g_strdup ("HOME"), g_strdup (pwent->pw_dir));
-                        g_hash_table_insert (hash, g_strdup ("PWD"), g_strdup (pwent->pw_dir));
+                pwent = getpwnam (username);
+                if (pwent != NULL) {
+                        if (pwent->pw_dir != NULL && pwent->pw_dir[0] != '\0') {
+                                g_hash_table_insert (hash, g_strdup ("HOME"),
+                                                     g_strdup (pwent->pw_dir));
+                                g_hash_table_insert (hash, g_strdup ("PWD"),
+                                                     g_strdup (pwent->pw_dir));
+                        }
+
+                        g_hash_table_insert (hash, g_strdup ("SHELL"),
+                                             g_strdup (pwent->pw_shell));
                 }
-
-                g_hash_table_insert (hash, g_strdup ("SHELL"), g_strdup (pwent->pw_shell));
         }
 
 #if 0
@@ -262,44 +268,32 @@ gdm_slave_run_script (GdmSlave   *slave,
         g_assert (login != NULL);
 
         script = g_build_filename (dir, slave->priv->display_name, NULL);
-        if (g_access (script, R_OK | X_OK) != 0) {
+        g_debug ("GdmSlave: Trying script %s", script);
+        if (! (g_file_test (script, G_FILE_TEST_IS_REGULAR)
+               && g_file_test (script, G_FILE_TEST_IS_EXECUTABLE))) {
                 g_debug ("GdmSlave: script %s not found; skipping", script);
                 g_free (script);
                 script = NULL;
         }
 
-        if (script == NULL &&
-            slave->priv->display_hostname != NULL) {
+        if (script == NULL
+            && slave->priv->display_hostname != NULL
+            && slave->priv->display_hostname[0] != '\0') {
                 script = g_build_filename (dir, slave->priv->display_hostname, NULL);
-                if (g_access (script, R_OK | X_OK) != 0) {
+                g_debug ("GdmSlave: Trying script %s", script);
+                if (! (g_file_test (script, G_FILE_TEST_IS_REGULAR)
+                       && g_file_test (script, G_FILE_TEST_IS_EXECUTABLE))) {
                         g_debug ("GdmSlave: script %s not found; skipping", script);
                         g_free (script);
                         script = NULL;
                 }
         }
 
-#if 0
-        if (script == NULL &&
-            SERVER_IS_XDMCP (d)) {
-                script = g_build_filename (dir, "XDMCP", NULL);
-                if (g_access (script, R_OK | X_OK) != 0) {
-                        g_free (script);
-                        script = NULL;
-                }
-        }
-        if (script == NULL &&
-            SERVER_IS_FLEXI (d)) {
-                script = g_build_filename (dir, "Flexi", NULL);
-                if (g_access (script, R_OK | X_OK) != 0) {
-                        g_free (script);
-                        script = NULL;
-                }
-        }
-#endif
-
         if (script == NULL) {
                 script = g_build_filename (dir, "Default", NULL);
-                if (g_access (script, R_OK | X_OK) != 0) {
+                g_debug ("GdmSlave: Trying script %s", script);
+                if (! (g_file_test (script, G_FILE_TEST_IS_REGULAR)
+                       && g_file_test (script, G_FILE_TEST_IS_EXECUTABLE))) {
                         g_debug ("GdmSlave: script %s not found; skipping", script);
                         g_free (script);
                         script = NULL;
@@ -393,6 +387,12 @@ gdm_slave_connect_to_x11_display (GdmSlave *slave)
         sigaddset (&mask, SIGCHLD);
         sigprocmask (SIG_BLOCK, &mask, &omask);
 
+        /* Give slave access to the display independent of current hostname */
+        XSetAuthorization ("MIT-MAGIC-COOKIE-1",
+                           strlen ("MIT-MAGIC-COOKIE-1"),
+                           slave->priv->display_x11_cookie->data,
+                           slave->priv->display_x11_cookie->len);
+
         slave->priv->server_display = XOpenDisplay (slave->priv->display_name);
 
         sigprocmask (SIG_SETMASK, &omask, NULL);
@@ -402,8 +402,35 @@ gdm_slave_connect_to_x11_display (GdmSlave *slave)
                 g_warning ("Unable to connect to display %s", slave->priv->display_name);
                 ret = FALSE;
         } else {
+                XHostAddress host_entries[2] = {
+                        { FamilyServerInterpreted },
+                        { FamilyServerInterpreted }
+                };
+                XServerInterpretedAddress si_entries[2];
+
                 g_debug ("GdmSlave: Connected to display %s", slave->priv->display_name);
                 ret = TRUE;
+
+                /* Give programs run by the slave and greeter access to the display
+                 * independent of current hostname
+                 */
+                si_entries[0].type = "localuser";
+                si_entries[0].typelength = strlen ("localuser");
+                si_entries[1].type = "localuser";
+                si_entries[1].typelength = strlen ("localuser");
+
+                si_entries[0].value = "root";
+                si_entries[0].valuelength = strlen ("root");
+                si_entries[1].value = GDM_USERNAME;
+                si_entries[1].valuelength = strlen (GDM_USERNAME);
+
+                host_entries[0].address = (char *) &si_entries[0];
+                host_entries[0].length = sizeof (XServerInterpretedAddress);
+                host_entries[1].address = (char *) &si_entries[1];
+                host_entries[1].length = sizeof (XServerInterpretedAddress);
+
+                XAddHosts (slave->priv->server_display, host_entries,
+                           G_N_ELEMENTS (host_entries));
         }
 
         return ret;
@@ -581,6 +608,25 @@ gdm_slave_real_start (GdmSlave *slave)
                 return FALSE;
         }
 
+        error = NULL;
+        res = dbus_g_proxy_call (slave->priv->display_proxy,
+                                 "GetX11Cookie",
+                                 &error,
+                                 G_TYPE_INVALID,
+                                 dbus_g_type_get_collection ("GArray", G_TYPE_CHAR),
+                                 &slave->priv->display_x11_cookie,
+                                 G_TYPE_INVALID);
+        if (! res) {
+                if (error != NULL) {
+                        g_warning ("Failed to get value: %s", error->message);
+                        g_error_free (error);
+                } else {
+                        g_warning ("Failed to get value");
+                }
+
+                return FALSE;
+        }
+
         error = NULL;
         res = dbus_g_proxy_call (slave->priv->display_proxy,
                                  "GetX11AuthorityFile",
@@ -717,6 +763,152 @@ gdm_slave_add_user_authorization (GdmSlave   *slave,
         return res;
 }
 
+static gchar *
+gdm_slave_parse_enriched_login (GdmSlave *slave,
+                                char *username,
+                                char *display_name)
+{
+        char in_buffer[20];
+        char **argv = NULL;
+        gint pipe1[2], in_buffer_len;
+        int  username_length;
+        pid_t pid;
+
+        if (username == NULL)
+                return (NULL);
+
+        /* A script may be used to generate the automatic/timed login name
+           based on the display/host by ending the name with the pipe symbol
+           '|'. */
+
+        username_length = strlen (username);
+        if (username_length > 0 && username[username_length-1] == '|') {
+                GPtrArray *env;
+                GError    *error;
+                gboolean   res;
+                char     **argv;
+                char      *std_output;
+                char      *std_error;
+                int       *exit_status;
+
+                /* Remove the pipe symbol */
+                username[username_length-1] = '\0';
+
+                argv = NULL;
+                error = NULL;
+                if (! g_shell_parse_argv (username, NULL, &argv, &error)) {
+                      g_warning ("Could not parse command: %s", error->message);
+                      return (NULL);
+                }
+
+                g_debug ("Calling script %s to acquire auto/timed username",
+                         username);
+
+                env = get_script_environment (slave, NULL);
+                error = NULL;
+                res = g_spawn_sync (NULL,
+                                    argv,
+                                    (char **)env->pdata,
+                                    G_SPAWN_SEARCH_PATH,
+                                    NULL,
+                                    NULL,
+                                    &std_output,
+                                    &std_error,
+                                    exit_status,
+                                    &error);
+
+                g_ptr_array_foreach (env, (GFunc)g_free, NULL);
+                g_ptr_array_free (env, TRUE);
+                g_strfreev (argv);
+
+                if (! res) {
+                        g_warning ("Unable to launch auto/timed login script: %s", error->message);
+                        g_error_free (error);
+	        } else {
+                        if (std_output != NULL) {
+                                g_strchomp (std_output);
+                                if (std_output[0] != '\0') {
+                                        return (g_strdup (std_output));
+                                }
+                        }
+                        return NULL;
+                }
+        } else {
+                return (g_strdup (username));
+        }
+}
+
+gboolean
+gdm_slave_get_timed_login_details (GdmSlave   *slave,
+                                   gboolean   *enabledp,
+                                   char      **usernamep,
+                                   int        *delayp)
+{
+        struct passwd *pwent;
+        GError        *error;
+        gboolean       res;
+        gboolean       enabled;
+        char          *username;
+        int            delay;
+
+        username = NULL;
+        enabled = FALSE;
+        delay = 0;
+
+        g_debug ("GdmSlave: Requesting timed login details");
+
+        error = NULL;
+        res = dbus_g_proxy_call (slave->priv->display_proxy,
+                                 "GetTimedLoginDetails",
+                                 &error,
+                                 G_TYPE_INVALID,
+                                 G_TYPE_BOOLEAN, &enabled,
+                                 G_TYPE_STRING, &username,
+                                 G_TYPE_INT, &delay,
+                                 G_TYPE_INVALID);
+
+        if (! res) {
+                if (error != NULL) {
+                        g_warning ("Failed to get timed login details: %s", error->message);
+                        g_error_free (error);
+                } else {
+                        g_warning ("Failed to get timed login details");
+                }
+        } else {
+                g_debug ("GdmSlave: Got timed login details: %d %s %d", enabled, username, delay);
+        }
+
+        if (usernamep != NULL) {
+                *usernamep = gdm_slave_parse_enriched_login (slave,
+                        username, slave->priv->display_name);
+        }
+        g_free (username);
+
+        if (usernamep != NULL && *usernamep != NULL) {
+                pwent = getpwnam (*usernamep);
+                if (pwent == NULL) {
+                        g_debug ("Invalid username %s for auto/timed login",
+                                 *usernamep);
+                        g_free (*usernamep);
+                        *usernamep = NULL;
+                } else {
+                        g_debug ("Using username %s for auto/timed login",
+                                 *usernamep);
+
+                        if (enabledp != NULL) {
+                                *enabledp = enabled;
+                        }
+                        if (delayp != NULL) {
+                                *delayp = delay;
+                        }
+               }
+        } else {
+                g_debug ("Invalid NULL username for auto/timed login");
+        }
+
+        return res;
+}
+
 static gboolean
 _get_uid_and_gid_for_user (const char *username,
                            uid_t      *uid,
@@ -841,9 +1033,9 @@ x11_session_is_on_seat (GdmSlave        *slave,
         return ret;
 }
 
-static char *
-_get_primary_user_session_id (GdmSlave   *slave,
-                              const char *username)
+char *
+gdm_slave_get_primary_session_id_for_user (GdmSlave   *slave,
+                                           const char *username)
 {
         gboolean    res;
         gboolean    can_activate_sessions;
@@ -954,6 +1146,7 @@ activate_session_id (GdmSlave   *slave,
         gboolean     ret;
 
         ret = FALSE;
+        reply = NULL;
 
         dbus_error_init (&local_error);
         message = dbus_message_new_method_call ("org.freedesktop.ConsoleKit",
@@ -986,6 +1179,13 @@ activate_session_id (GdmSlave   *slave,
 
         ret = TRUE;
  out:
+        if (message != NULL) {
+                dbus_message_unref (message);
+        }
+        if (reply != NULL) {
+                dbus_message_unref (reply);
+        }
+
         return ret;
 }
 
@@ -1012,7 +1212,9 @@ session_unlock (GdmSlave   *slave,
                                                            message,
                                                            -1, &error);
         dbus_message_unref (message);
-        dbus_message_unref (reply);
+        if (reply != NULL) {
+                dbus_message_unref (reply);
+        }
         dbus_connection_flush (dbus_g_connection_get_connection (slave->priv->connection));
 
         if (dbus_error_is_set (&error)) {
@@ -1033,7 +1235,7 @@ gdm_slave_switch_to_user_session (GdmSlave   *slave,
 
         ret = FALSE;
 
-        ssid_to_activate = _get_primary_user_session_id (slave, username);
+        ssid_to_activate = gdm_slave_get_primary_session_id_for_user (slave, username);
         if (ssid_to_activate == NULL) {
                 g_debug ("GdmSlave: unable to determine session to activate");
                 goto out;
@@ -1347,7 +1549,15 @@ gdm_slave_finalize (GObject *object)
 
         gdm_slave_real_stop (slave);
 
+        g_free (slave->priv->id);
         g_free (slave->priv->display_id);
+        g_free (slave->priv->display_name);
+        g_free (slave->priv->display_hostname);
+        g_free (slave->priv->display_seat_id);
+        g_free (slave->priv->display_x11_authority_file);
+        g_free (slave->priv->parent_display_name);
+        g_free (slave->priv->parent_display_x11_authority_file);
+        g_array_free (slave->priv->display_x11_cookie, TRUE);
 
         G_OBJECT_CLASS (gdm_slave_parent_class)->finalize (object);
 }
