@@ -30,10 +30,15 @@
 #include <sys/stat.h>
 #include <sys/types.h>
 
+#ifdef HAVE_PATHS_H
+#include <paths.h>
+#endif /* HAVE_PATHS_H */
+
 #include <glib.h>
 #include <glib/gi18n.h>
 #include <glib/gstdio.h>
 #include <glib-object.h>
+#include <gio/gio.h>
 
 #include <dbus/dbus.h>
 #include <dbus/dbus-glib.h>
@@ -58,6 +63,17 @@
 #define DEFAULT_MAX_ICON_SIZE   128
 #define DEFAULT_USER_MAX_FILE   65536
 
+#ifdef __sun
+#define DEFAULT_MINIMAL_UID     100
+#else
+#define DEFAULT_MINIMAL_UID     500
+#endif
+
+#ifndef _PATH_SHELLS
+#define _PATH_SHELLS    "/etc/shells"
+#endif
+#define PATH_PASSWD     "/etc/passwd"
+
 #define DEFAULT_GLOBAL_FACE_DIR DATADIR "/faces"
 #define DEFAULT_USER_ICON       "stock_person"
 #define DEFAULT_EXCLUDE         { "bin",        \
@@ -73,7 +89,9 @@
                                   "uucp",       \
                                   "operator",   \
                                   "nobody",     \
-                                  "gdm",        \
+                                  "nobody4",    \
+                                  "noaccess",   \
+                                  GDM_USERNAME, \
                                   "postgres",   \
                                   "pvm",        \
                                   "rpm",        \
@@ -86,10 +104,14 @@ struct GdmUserManagerPrivate
         GHashTable            *users;
         GHashTable            *sessions;
         GHashTable            *exclusions;
+        GHashTable            *shells;
         DBusGConnection       *connection;
         DBusGProxy            *seat_proxy;
         char                  *seat_id;
 
+        GFileMonitor          *passwd_monitor;
+        GFileMonitor          *shells_monitor;
+
         guint                  reload_id;
         guint                  ck_history_id;
 
@@ -97,6 +119,8 @@ struct GdmUserManagerPrivate
 };
 
 enum {
+        LOADING_USERS,
+        USERS_LOADED,
         USER_ADDED,
         USER_REMOVED,
         USER_IS_LOGGED_IN_CHANGED,
@@ -133,8 +157,12 @@ start_new_login_session (GdmUserManager *manager)
 
         res = g_spawn_command_line_async ("gdmflexiserver -s", &error);
         if (! res) {
-                g_warning ("Unable to start new login: %s", error->message);
-                g_error_free (error);
+                if (error != NULL) {
+                        g_warning ("Unable to start new login: %s", error->message);
+                        g_error_free (error);
+                } else {
+                        g_warning ("Unable to start new login");
+                }
         }
 
         return res;
@@ -170,9 +198,13 @@ _get_primary_user_session_id (GdmUserManager *manager,
                                  G_TYPE_BOOLEAN, &can_activate_sessions,
                                  G_TYPE_INVALID);
         if (! res) {
-                g_warning ("unable to determine if seat can activate sessions: %s",
-                           error->message);
-                g_error_free (error);
+                if (error != NULL) {
+                        g_warning ("unable to determine if seat can activate sessions: %s",
+                                   error->message);
+                        g_error_free (error);
+                } else {
+                        g_warning ("unable to determine if seat can activate sessions");
+                }
                 goto out;
         }
 
@@ -216,6 +248,7 @@ activate_session_id (GdmUserManager *manager,
         gboolean     ret;
 
         ret = FALSE;
+        reply = NULL;
 
         dbus_error_init (&local_error);
         message = dbus_message_new_method_call ("org.freedesktop.ConsoleKit",
@@ -248,6 +281,13 @@ activate_session_id (GdmUserManager *manager,
 
         ret = TRUE;
  out:
+        if (message != NULL) {
+                dbus_message_unref (message);
+        }
+        if (reply != NULL) {
+                dbus_message_unref (reply);
+        }
+
         return ret;
 }
 
@@ -281,8 +321,12 @@ session_is_login_window (GdmUserManager *manager,
                                  G_TYPE_STRING, &session_type,
                                  G_TYPE_INVALID);
         if (! res) {
-                g_debug ("Failed to identify the session type: %s", error->message);
-                g_error_free (error);
+                if (error != NULL) {
+                        g_debug ("Failed to identify the session type: %s", error->message);
+                        g_error_free (error);
+                } else {
+                        g_debug ("Failed to identify the session type");
+                }
                 goto out;
         }
 
@@ -328,9 +372,13 @@ _get_login_window_session_id (GdmUserManager *manager)
                                  G_TYPE_BOOLEAN, &can_activate_sessions,
                                  G_TYPE_INVALID);
         if (! res) {
-                g_warning ("unable to determine if seat can activate sessions: %s",
-                           error->message);
-                g_error_free (error);
+                if (error != NULL) {
+                        g_warning ("unable to determine if seat can activate sessions: %s",
+                                   error->message);
+                        g_error_free (error);
+                } else {
+                        g_warning ("unable to determine if seat can activate sessions");
+                }
                 goto out;
         }
 
@@ -347,9 +395,13 @@ _get_login_window_session_id (GdmUserManager *manager)
                                  dbus_g_type_get_collection ("GPtrArray", DBUS_TYPE_G_OBJECT_PATH), &sessions,
                                  G_TYPE_INVALID);
         if (! res) {
-                g_warning ("unable to determine sessions for user: %s",
-                           error->message);
-                g_error_free (error);
+                if (error != NULL) {
+                        g_warning ("unable to determine sessions for user: %s",
+                                   error->message);
+                        g_error_free (error);
+                } else {
+                        g_warning ("unable to determine sessions for user");
+                }
                 goto out;
         }
 
@@ -476,7 +528,7 @@ get_seat_id_for_session (DBusGConnection *connection,
                                            session_id,
                                            CK_SESSION_INTERFACE);
         if (proxy == NULL) {
-                g_warning ("Failed to connect to the ConsoleKit seat object");
+                g_warning ("Failed to connect to the ConsoleKit session object");
                 goto out;
         }
 
@@ -488,8 +540,12 @@ get_seat_id_for_session (DBusGConnection *connection,
                                  DBUS_TYPE_G_OBJECT_PATH, &seat_id,
                                  G_TYPE_INVALID);
         if (! res) {
-                g_debug ("Failed to identify the current seat: %s", error->message);
-                g_error_free (error);
+                if (error != NULL) {
+                        g_debug ("Failed to identify the current seat: %s", error->message);
+                        g_error_free (error);
+                } else {
+                        g_debug ("Failed to identify the current seat");
+                }
         }
  out:
         if (proxy != NULL) {
@@ -516,7 +572,7 @@ get_x11_display_for_session (DBusGConnection *connection,
                                            session_id,
                                            CK_SESSION_INTERFACE);
         if (proxy == NULL) {
-                g_warning ("Failed to connect to the ConsoleKit seat object");
+                g_warning ("Failed to connect to the ConsoleKit session object");
                 goto out;
         }
 
@@ -528,8 +584,12 @@ get_x11_display_for_session (DBusGConnection *connection,
                                  G_TYPE_STRING, &x11_display,
                                  G_TYPE_INVALID);
         if (! res) {
-                g_debug ("Failed to identify the x11 display: %s", error->message);
-                g_error_free (error);
+                if (error != NULL) {
+                        g_debug ("Failed to identify the x11 display: %s", error->message);
+                        g_error_free (error);
+                } else {
+                        g_debug ("Failed to identify the x11 display");
+                }
         }
  out:
         if (proxy != NULL) {
@@ -623,8 +683,12 @@ add_sessions_for_user (GdmUserManager *manager,
                                  &sessions,
                                  G_TYPE_INVALID);
         if (! res) {
-                g_debug ("Failed to find sessions for user: %s", error->message);
-                g_error_free (error);
+                if (error != NULL) {
+                        g_debug ("Failed to find sessions for user: %s", error->message);
+                        g_error_free (error);
+                } else {
+                        g_debug ("Failed to find sessions for user");
+                }
                 goto out;
         }
 
@@ -637,6 +701,7 @@ add_sessions_for_user (GdmUserManager *manager,
                 maybe_add_session_for_user (manager, user, ssid);
         }
 
+        g_ptr_array_foreach (sessions, (GFunc)g_free, NULL);
         g_ptr_array_free (sessions, TRUE);
 
  out:
@@ -662,11 +727,39 @@ create_user (GdmUserManager *manager)
         return user;
 }
 
+static gint
+match_real_name_cmpfunc (gconstpointer a,
+                         gconstpointer b)
+{
+        if (a == b)
+                return -1;
+
+        return g_strcmp0 (gdm_user_get_real_name ((GdmUser *) a),
+                          gdm_user_get_real_name ((GdmUser *) b));
+}
+
+static gboolean
+match_real_name_hrfunc (gpointer key,
+                        gpointer value,
+                        gpointer user_data)
+{
+        return (g_strcmp0 (user_data, gdm_user_get_real_name (value)) == 0);
+}
+
 static void
 add_user (GdmUserManager *manager,
           GdmUser        *user)
 {
+        GdmUser *dup;
+
         add_sessions_for_user (manager, user);
+        dup = g_hash_table_find (manager->priv->users,
+                                 match_real_name_hrfunc,
+                                 (char *) gdm_user_get_real_name (user));
+        if (dup != NULL) {
+                _gdm_user_show_full_display_name (user);
+                _gdm_user_show_full_display_name (dup);
+        }
         g_hash_table_insert (manager->priv->users,
                              g_strdup (gdm_user_get_user_name (user)),
                              g_object_ref (user));
@@ -721,8 +814,12 @@ get_current_seat_id (DBusGConnection *connection)
                                  &session_id,
                                  G_TYPE_INVALID);
         if (! res) {
-                g_debug ("Failed to identify the current session: %s", error->message);
-                g_error_free (error);
+                if (error != NULL) {
+                        g_debug ("Failed to identify the current session: %s", error->message);
+                        g_error_free (error);
+                } else {
+                        g_debug ("Failed to identify the current session");
+                }
                 goto out;
         }
 
@@ -744,7 +841,7 @@ get_uid_from_session_id (GdmUserManager *manager,
 {
         DBusGProxy      *proxy;
         GError          *error;
-        int              uid;
+        guint            uid;
         gboolean         res;
 
         proxy = dbus_g_proxy_new_for_name (manager->priv->connection,
@@ -761,13 +858,17 @@ get_uid_from_session_id (GdmUserManager *manager,
                                  "GetUnixUser",
                                  &error,
                                  G_TYPE_INVALID,
-                                 G_TYPE_INT, &uid,
+                                 G_TYPE_UINT, &uid,
                                  G_TYPE_INVALID);
         g_object_unref (proxy);
 
         if (! res) {
-                g_warning ("Failed to query the session: %s", error->message);
-                g_error_free (error);
+                if (error != NULL) {
+                        g_warning ("Failed to query the session: %s", error->message);
+                        g_error_free (error);
+                } else {
+                        g_warning ("Failed to query the session");
+                }
                 return FALSE;
         }
 
@@ -881,8 +982,12 @@ get_seat_proxy (GdmUserManager *manager)
         error = NULL;
         manager->priv->connection = dbus_g_bus_get (DBUS_BUS_SYSTEM, &error);
         if (manager->priv->connection == NULL) {
-                g_warning ("Failed to connect to the D-Bus daemon: %s", error->message);
-                g_error_free (error);
+                if (error != NULL) {
+                        g_warning ("Failed to connect to the D-Bus daemon: %s", error->message);
+                        g_error_free (error);
+                } else {
+                        g_warning ("Failed to connect to the D-Bus daemon");
+                }
                 return;
         }
 
@@ -901,8 +1006,13 @@ get_seat_proxy (GdmUserManager *manager)
                                                  &error);
 
         if (proxy == NULL) {
-                g_warning ("Failed to connect to the ConsoleKit seat object: %s", error->message);
-                g_error_free (error);
+                if (error != NULL) {
+                        g_warning ("Failed to connect to the ConsoleKit seat object: %s",
+                                   error->message);
+                        g_error_free (error);
+                } else {
+                        g_warning ("Failed to connect to the ConsoleKit seat object");
+                }
                 return;
         }
 
@@ -910,11 +1020,11 @@ get_seat_proxy (GdmUserManager *manager)
 
         dbus_g_proxy_add_signal (proxy,
                                  "SessionAdded",
-                                 G_TYPE_STRING,
+                                 DBUS_TYPE_G_OBJECT_PATH,
                                  G_TYPE_INVALID);
         dbus_g_proxy_add_signal (proxy,
                                  "SessionRemoved",
-                                 G_TYPE_STRING,
+                                 DBUS_TYPE_G_OBJECT_PATH,
                                  G_TYPE_INVALID);
         dbus_g_proxy_connect_signal (proxy,
                                      "SessionAdded",
@@ -1054,7 +1164,11 @@ parse_ck_history_line (const char *line,
         error = NULL;
         re = g_regex_new ("(?P<username>[0-9a-zA-Z]+)[ ]+(?P<frequency>[0-9]+)", 0, 0, &error);
         if (re == NULL) {
-                g_critical (error->message);
+                if (error != NULL) {
+                        g_critical ("%s", error->message);
+                } else {
+                        g_critical ("Error in regex call");
+                }
                 goto out;
         }
 
@@ -1115,8 +1229,15 @@ process_ck_history_line (GdmUserManager *manager,
         }
 
         user = gdm_user_manager_get_user (manager, username);
+        if (user == NULL) {
+                g_debug ("GdmUserManager: unable to lookup user '%s'", username);
+                g_free (username);
+                return;
+        }
+
         g_object_set (user, "login-frequency", frequency, NULL);
         g_signal_emit (manager, signals [USER_LOGIN_FREQUENCY_CHANGED], 0, user);
+        g_free (username);
 }
 
 static gboolean
@@ -1153,6 +1274,8 @@ ck_history_watch (GIOChannel     *source,
         }
 
         if (done) {
+                g_signal_emit (G_OBJECT (manager), signals[USERS_LOADED], 0);
+
                 manager->priv->ck_history_id = 0;
                 return FALSE;
         }
@@ -1161,7 +1284,7 @@ ck_history_watch (GIOChannel     *source,
 }
 
 static void
-reload_users (GdmUserManager *manager)
+reload_ck_history (GdmUserManager *manager)
 {
         char       *command;
         const char *seat_id;
@@ -1188,8 +1311,12 @@ reload_users (GdmUserManager *manager)
         g_debug ("GdmUserManager: running '%s'", command);
         error = NULL;
         if (! g_shell_parse_argv (command, NULL, &argv, &error)) {
-                g_warning ("Could not parse command: %s", error->message);
-                g_error_free (error);
+                if (error != NULL) {
+                        g_warning ("Could not parse command: %s", error->message);
+                        g_error_free (error);
+                } else {
+                        g_warning ("Could not parse command");
+                }
                 goto out;
         }
 
@@ -1207,8 +1334,12 @@ reload_users (GdmUserManager *manager)
                                         &error);
         g_strfreev (argv);
         if (! res) {
-                g_warning ("Unable to run ck-history: %s", error->message);
-                g_error_free (error);
+                if (error != NULL) {
+                        g_warning ("Unable to run ck-history: %s", error->message);
+                        g_error_free (error);
+                } else {
+                        g_warning ("Unable to run ck-history");
+                }
                 goto out;
         }
 
@@ -1227,11 +1358,132 @@ reload_users (GdmUserManager *manager)
         g_free (command);
 }
 
+static void
+reload_passwd (GdmUserManager *manager)
+{
+        struct passwd *pwent;
+        GSList        *old_users;
+        GSList        *new_users;
+        GSList        *list;
+        GSList        *dup;
+        FILE          *fp;
+
+        old_users = NULL;
+        new_users = NULL;
+
+        errno = 0;
+        fp = fopen (PATH_PASSWD, "r");
+        if (fp == NULL) {
+                g_warning ("Unable to open %s: %s", PATH_PASSWD, g_strerror (errno));
+                goto out;
+        }
+
+        g_hash_table_foreach (manager->priv->users, listify_hash_values_hfunc, &old_users);
+        g_slist_foreach (old_users, (GFunc) g_object_ref, NULL);
+
+        /* Make sure we keep users who are logged in no matter what. */
+        for (list = old_users; list; list = list->next) {
+                if (gdm_user_get_num_sessions (list->data) > 0) {
+                        g_object_freeze_notify (G_OBJECT (list->data));
+                        _gdm_user_show_short_display_name (list->data);
+                        new_users = g_slist_prepend (new_users, g_object_ref (list->data));
+                }
+        }
+
+        for (pwent = fgetpwent (fp); pwent != NULL; pwent = fgetpwent (fp)) {
+                GdmUser *user;
+
+                user = NULL;
+
+                /* Skip users below MinimalUID... */
+                if (pwent->pw_uid < DEFAULT_MINIMAL_UID) {
+                        continue;
+                }
+
+                /* ...And users w/ invalid shells... */
+                if (pwent->pw_shell == NULL ||
+                    !g_hash_table_lookup (manager->priv->shells, pwent->pw_shell)) {
+                        g_debug ("GdmUserManager: skipping user with bad shell: %s", pwent->pw_name);
+                        continue;
+                }
+
+                /* ...And explicitly excluded users */
+                if (g_hash_table_lookup (manager->priv->exclusions, pwent->pw_name)) {
+                        g_debug ("GdmUserManager: explicitly skipping user: %s", pwent->pw_name);
+                        continue;
+                }
+
+                user = g_hash_table_lookup (manager->priv->users, pwent->pw_name);
+
+                /* Update users already in the *new* list */
+                if (g_slist_find (new_users, user)) {
+                        _gdm_user_update (user, pwent);
+                        continue;
+                }
+
+                if (user == NULL) {
+                        user = create_user (manager);
+                } else {
+                        g_object_ref (user);
+                }
+
+                /* Freeze & update users not already in the new list */
+                g_object_freeze_notify (G_OBJECT (user));
+                _gdm_user_update (user, pwent);
+
+                new_users = g_slist_prepend (new_users, user);
+        }
+
+        /* Go through and handle removed users */
+        for (list = old_users; list; list = list->next) {
+                if (! g_slist_find (new_users, list->data)) {
+                        g_signal_emit (manager, signals[USER_REMOVED], 0, list->data);
+                        g_hash_table_remove (manager->priv->users,
+                                             gdm_user_get_user_name (list->data));
+                }
+        }
+
+        /* Go through and handle added users or update display names */
+        for (list = new_users; list; list = list->next) {
+                if (g_slist_find (old_users, list->data)) {
+                        dup = g_slist_find_custom (new_users,
+                                                   list->data,
+                                                   match_real_name_cmpfunc);
+                        if (dup != NULL) {
+                                _gdm_user_show_full_display_name (list->data);
+                                _gdm_user_show_full_display_name (dup->data);
+                        }
+                } else {
+                        add_user (manager, list->data);
+                }
+        }
+
+ out:
+        /* Cleanup */
+
+        fclose (fp);
+
+        g_slist_foreach (new_users, (GFunc) g_object_thaw_notify, NULL);
+        g_slist_foreach (new_users, (GFunc) g_object_unref, NULL);
+        g_slist_free (new_users);
+
+        g_slist_foreach (old_users, (GFunc) g_object_unref, NULL);
+        g_slist_free (old_users);
+}
+
+static void
+reload_users (GdmUserManager *manager)
+{
+        reload_ck_history (manager);
+        reload_passwd (manager);
+}
+
 static gboolean
 reload_users_timeout (GdmUserManager *manager)
 {
         reload_users (manager);
         manager->priv->reload_id = 0;
+
         return FALSE;
 }
 
@@ -1242,9 +1494,65 @@ queue_reload_users (GdmUserManager *manager)
                 return;
         }
 
+        g_signal_emit (G_OBJECT (manager), signals[LOADING_USERS], 0);
         manager->priv->reload_id = g_idle_add ((GSourceFunc)reload_users_timeout, manager);
 }
 
+static void
+reload_shells (GdmUserManager *manager)
+{
+        char *shell;
+
+        setusershell ();
+
+        g_hash_table_remove_all (manager->priv->shells);
+        for (shell = getusershell (); shell != NULL; shell = getusershell ()) {
+                /* skip well known not-real shells */
+                if (shell == NULL
+                    || strcmp (shell, "/sbin/nologin") == 0
+                    || strcmp (shell, "/bin/false") == 0) {
+                        g_debug ("GdmUserManager: skipping shell %s", shell);
+                        continue;
+                }
+                g_hash_table_insert (manager->priv->shells,
+                                     g_strdup (shell),
+                                     GUINT_TO_POINTER (TRUE));
+        }
+
+        endusershell ();
+}
+
+static void
+on_shells_monitor_changed (GFileMonitor     *monitor,
+                           GFile            *file,
+                           GFile            *other_file,
+                           GFileMonitorEvent event_type,
+                           GdmUserManager   *manager)
+{
+        if (event_type != G_FILE_MONITOR_EVENT_CHANGED &&
+            event_type != G_FILE_MONITOR_EVENT_CREATED) {
+                return;
+        }
+
+        reload_shells (manager);
+        reload_passwd (manager);
+}
+
+static void
+on_passwd_monitor_changed (GFileMonitor     *monitor,
+                           GFile            *file,
+                           GFile            *other_file,
+                           GFileMonitorEvent event_type,
+                           GdmUserManager   *manager)
+{
+        if (event_type != G_FILE_MONITOR_EVENT_CHANGED &&
+            event_type != G_FILE_MONITOR_EVENT_CREATED) {
+                return;
+        }
+
+        reload_passwd (manager);
+}
+
 static void
 gdm_user_manager_class_init (GdmUserManagerClass *klass)
 {
@@ -1252,6 +1560,22 @@ gdm_user_manager_class_init (GdmUserManagerClass *klass)
 
         object_class->finalize = gdm_user_manager_finalize;
 
+        signals [LOADING_USERS] =
+                g_signal_new ("loading-users",
+                              G_TYPE_FROM_CLASS (klass),
+                              G_SIGNAL_RUN_LAST,
+                              G_STRUCT_OFFSET (GdmUserManagerClass, loading_users),
+                              NULL, NULL,
+                              g_cclosure_marshal_VOID__VOID,
+                              G_TYPE_NONE, 0);
+        signals [USERS_LOADED] =
+                g_signal_new ("users-loaded",
+                              G_TYPE_FROM_CLASS (klass),
+                              G_SIGNAL_RUN_LAST,
+                              G_STRUCT_OFFSET (GdmUserManagerClass, users_loaded),
+                              NULL, NULL,
+                              g_cclosure_marshal_VOID__VOID,
+                              G_TYPE_NONE, 0);
         signals [USER_ADDED] =
                 g_signal_new ("user-added",
                               G_TYPE_FROM_CLASS (klass),
@@ -1292,6 +1616,8 @@ static void
 gdm_user_manager_init (GdmUserManager *manager)
 {
         int            i;
+        GFile         *file;
+        GError        *error;
         const char    *exclude_default[] = DEFAULT_EXCLUDE;
 
         manager->priv = GDM_USER_MANAGER_GET_PRIVATE (manager);
@@ -1313,10 +1639,58 @@ gdm_user_manager_init (GdmUserManager *manager)
                                      GUINT_TO_POINTER (TRUE));
         }
 
+        /* /etc/shells */
+        manager->priv->shells = g_hash_table_new_full (g_str_hash,
+                                                       g_str_equal,
+                                                       g_free,
+                                                       NULL);
+        reload_shells (manager);
+        file = g_file_new_for_path (_PATH_SHELLS);
+        error = NULL;
+        manager->priv->shells_monitor = g_file_monitor_file (file,
+                                                             G_FILE_MONITOR_NONE,
+                                                             NULL,
+                                                             &error);
+        if (manager->priv->shells_monitor != NULL) {
+                g_signal_connect (manager->priv->shells_monitor,
+                                  "changed",
+                                  G_CALLBACK (on_shells_monitor_changed),
+                                  manager);
+        } else {
+                if (error != NULL) {
+                        g_warning ("Unable to monitor %s: %s", _PATH_SHELLS, error->message);
+                        g_error_free (error);
+                } else {
+                        g_warning ("Unable to monitor %s", _PATH_SHELLS);
+                }
+        }
+        g_object_unref (file);
+
+        /* /etc/passwd */
         manager->priv->users = g_hash_table_new_full (g_str_hash,
                                                       g_str_equal,
                                                       g_free,
                                                       (GDestroyNotify) g_object_run_dispose);
+        file = g_file_new_for_path (PATH_PASSWD);
+        manager->priv->passwd_monitor = g_file_monitor_file (file,
+                                                             G_FILE_MONITOR_NONE,
+                                                             NULL,
+                                                             &error);
+        if (manager->priv->passwd_monitor != NULL) {
+                g_signal_connect (manager->priv->passwd_monitor,
+                                  "changed",
+                                  G_CALLBACK (on_passwd_monitor_changed),
+                                  manager);
+        } else {
+                if (error != NULL) {
+                        g_warning ("Unable to monitor %s: %s", PATH_PASSWD, error->message);
+                        g_error_free (error);
+                } else {
+                        g_warning ("Unable to monitor %s", PATH_PASSWD);
+                }
+        }
+        g_object_unref (file);
+
 
         get_seat_proxy (manager);
 
@@ -1353,8 +1727,12 @@ gdm_user_manager_finalize (GObject *object)
 
         g_hash_table_destroy (manager->priv->sessions);
 
+        g_file_monitor_cancel (manager->priv->passwd_monitor);
         g_hash_table_destroy (manager->priv->users);
 
+        g_file_monitor_cancel (manager->priv->shells_monitor);
+        g_hash_table_destroy (manager->priv->shells);
+
         g_free (manager->priv->seat_id);
 
         G_OBJECT_CLASS (gdm_user_manager_parent_class)->finalize (object);
